@@ -1,5 +1,6 @@
 /* Handle types for the GNU compiler for the Java(TM) language.
-   Copyright (C) 1996, 97-98, 1999 Free Software Foundation, Inc.
+   Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001
+   Free Software Foundation, Inc.
 
 This file is part of GNU CC.
 
@@ -33,6 +34,13 @@ The Free Software Foundation is independent of Sun Microsystems, Inc.  */
 #include "jcf.h"
 #include "convert.h"
 #include "toplev.h"
+#include "ggc.h"
+
+static tree convert_ieee_real_to_integer PARAMS ((tree, tree));
+static tree parse_signature_type PARAMS ((const unsigned char **,
+					 const unsigned char *));
+static tree lookup_do PARAMS ((tree, tree, tree, tree, tree (*)(tree)));
+static tree build_null_signature PARAMS ((tree));
 
 tree * type_map;
 extern struct obstack permanent_obstack;
@@ -46,8 +54,10 @@ set_local_type (slot, type)
 {
   int max_locals = DECL_MAX_LOCALS(current_function_decl);
   int nslots = TYPE_IS_WIDE (type) ? 2 : 1;
+
   if (slot < 0 || slot + nslots - 1 >= max_locals)
-    fatal ("invalid local variable index");
+    abort ();
+
   type_map[slot] = type;
   while (--nslots > 0)
     type_map[++slot] = void_type_node;
@@ -55,23 +65,43 @@ set_local_type (slot, type)
 
 /* Convert an IEEE real to an integer type.  The result of such a
    conversion when the source operand is a NaN isn't defined by
-   IEEE754, but by the Java language standard: it must be zero.  This
-   conversion produces something like:
-   
-   ({ double tmp = expr; (tmp != tmp) ? 0 : (int)tmp; })
+   IEEE754, but by the Java language standard: it must be zero.  Also,
+   overflows must be clipped to within range.  This conversion
+   produces something like:
 
-   */
+      ((expr >= (float)MAX_INT)
+       ? MAX_INT 
+       : ((expr <= (float)MIN_INT)
+	  ? MIN_INT
+	  : ((expr != expr)
+	     ? 0 
+	     : (int)expr))) */
 
 static tree
 convert_ieee_real_to_integer (type, expr)
      tree type, expr;
 {
+  tree result;
   expr = save_expr (expr);
 
-  return build (COND_EXPR, type, 
-		build (NE_EXPR, boolean_type_node, expr, expr),
-		convert (type, integer_zero_node),
-		convert_to_integer (type, expr));
+  result = build (COND_EXPR, type,
+		  build (NE_EXPR, boolean_type_node, expr, expr),
+		  convert (type, integer_zero_node),
+		  convert_to_integer (type, expr));
+		  
+  result = build (COND_EXPR, type, 
+		  build (LE_EXPR, boolean_type_node, expr, 
+			 convert (TREE_TYPE (expr), TYPE_MIN_VALUE (type))),
+		  TYPE_MIN_VALUE (type),
+		  result);
+
+  result = build (COND_EXPR, type,
+		  build (GE_EXPR, boolean_type_node, expr, 
+			 convert (TREE_TYPE (expr), TYPE_MAX_VALUE (type))),	
+		  TYPE_MAX_VALUE (type),
+		  result);
+
+  return result;
 }  
 
 /* Create an expression whose value is that of EXPR,
@@ -86,6 +116,9 @@ convert (type, expr)
 {
   register enum tree_code code = TREE_CODE (type);
 
+  if (!expr)
+   return error_mark_node;
+
   if (do_not_fold)
     return build1 (NOP_EXPR, type, expr);
 
@@ -100,12 +133,9 @@ convert (type, expr)
     return fold (convert_to_boolean (type, expr));
   if (code == INTEGER_TYPE)
     {
-      if (TREE_CODE (TREE_TYPE (expr)) == REAL_TYPE
-#ifdef TARGET_SOFT_FLOAT
-	  && !TARGET_SOFT_FLOAT
-#endif
-	  && !flag_emit_class_files
-	  && !flag_fast_math
+      if (! flag_unsafe_math_optimizations
+	  && ! flag_emit_class_files
+	  && TREE_CODE (TREE_TYPE (expr)) == REAL_TYPE
 	  && TARGET_FLOAT_FORMAT == IEEE_FLOAT_FORMAT)
 	return fold (convert_ieee_real_to_integer (type, expr));
       else
@@ -323,21 +353,35 @@ java_array_type_length (array_type)
   if (arfld != NULL_TREE)
     {
       tree index_type = TYPE_DOMAIN (TREE_TYPE (arfld));
-      tree high = TYPE_MAX_VALUE (index_type);
-      if (TREE_CODE (high) == INTEGER_CST)
-	return TREE_INT_CST_LOW (high) + 1;
+      if (index_type != NULL_TREE)
+	{
+	  tree high = TYPE_MAX_VALUE (index_type);
+	  if (TREE_CODE (high) == INTEGER_CST)
+	    return TREE_INT_CST_LOW (high) + 1;
+	}
     }
   return -1;
 }
 
+/* An array of unknown length will be ultimately given an length of
+   -2, so that we can still have `length' producing a negative value
+   even if found. This was part of an optimization amaing at removing
+   `length' from static arrays. We could restore it, FIXME.  */
+
 tree
 build_prim_array_type (element_type, length)
      tree element_type;
      HOST_WIDE_INT length;
 {
-  tree max_index = build_int_2 (length - 1, 0);
-  TREE_TYPE (max_index) = sizetype;
-  return build_array_type (element_type, build_index_type (max_index));
+  tree index = NULL;
+
+  if (length != -1)
+    {
+      tree max_index = build_int_2 (length - 1, (0 == length ? -1 : 0));
+      TREE_TYPE (max_index) = sizetype;
+      index = build_index_type (max_index);
+    }
+  return build_array_type (element_type, index);
 }
 
 /* Return a Java array type with a given ELEMENT_TYPE and LENGTH.
@@ -349,11 +393,15 @@ build_java_array_type (element_type, length)
      tree element_type;
      HOST_WIDE_INT length;
 {
-  tree sig, t, fld;
+  tree sig, t, fld, atype, arfld;
   char buf[12];
   tree elsig = build_java_signature (element_type);
   tree el_name = element_type;
-  sprintf (buf, length >= 0 ? "[%d" : "[", length);
+  buf[0] = '[';
+  if (length >= 0)
+    sprintf (buf+1, HOST_WIDE_INT_PRINT_DEC, length);
+  else
+    buf[1] = '\0';
   sig = ident_subst (IDENTIFIER_POINTER (elsig), IDENTIFIER_LENGTH (elsig),
 		     buf, 0, 0, "");
   t = IDENTIFIER_SIGNATURE_TYPE (sig);
@@ -368,7 +416,9 @@ build_java_array_type (element_type, length)
   el_name = TYPE_NAME (el_name);
   if (TREE_CODE (el_name) == TYPE_DECL)
     el_name = DECL_NAME (el_name);
-  TYPE_NAME (t) = identifier_subst (el_name, "", '.', '.', "[]");
+  TYPE_NAME (t) = build_decl (TYPE_DECL,
+                             identifier_subst (el_name, "", '.', '.', "[]"),
+                             t);
 
   set_java_signature (t, sig);
   set_super_info (0, t, object_type_node, 0);
@@ -377,23 +427,18 @@ build_java_array_type (element_type, length)
   TYPE_ARRAY_ELEMENT (t) = element_type;
 
   /* Add length pseudo-field. */
-  push_obstacks (&permanent_obstack, &permanent_obstack);
   fld = build_decl (FIELD_DECL, get_identifier ("length"), int_type_node);
   TYPE_FIELDS (t) = fld;
   DECL_CONTEXT (fld) = t;
   FIELD_PUBLIC (fld) = 1;
   FIELD_FINAL (fld) = 1;
+  TREE_READONLY (fld) = 1;
 
-  if (length >= 0)
-    {
-      tree atype = build_prim_array_type (element_type, length);
-      tree arfld = build_decl (FIELD_DECL, get_identifier ("data"), atype);
-      DECL_CONTEXT (arfld) = t;
-      TREE_CHAIN (fld) = arfld;
-    }
-  else
-    TYPE_ALIGN (t) = TYPE_ALIGN (element_type);
-  pop_obstacks ();
+  atype = build_prim_array_type (element_type, length);
+  arfld = build_decl (FIELD_DECL, get_identifier ("data"), atype);
+  DECL_CONTEXT (arfld) = t;
+  TREE_CHAIN (fld) = arfld;
+  DECL_ALIGN (arfld) = TYPE_ALIGN (element_type);
 
   /* We could layout_class, but that loads java.lang.Object prematurely.
    * This is called by the parser, and it is a bad idea to do load_class
@@ -446,9 +491,11 @@ parse_signature_type (ptr, limit)
      const unsigned char **ptr, *limit;
 {
   tree type;
-  if ((*ptr) >= limit)
-    fatal ("bad signature string");
-  switch (*(*ptr))
+
+  if (*ptr >= limit)
+    abort ();
+
+  switch (**ptr)
     {
     case 'B':  (*ptr)++;  return byte_type_node;
     case 'C':  (*ptr)++;  return char_type_node;
@@ -471,7 +518,7 @@ parse_signature_type (ptr, limit)
 	for ( ; ; str++)
 	  {
 	    if (str >= limit)
-	      fatal ("bad signature string");
+	      abort ();
 	    if (*str == ';')
 	      break;
 	  }
@@ -480,7 +527,7 @@ parse_signature_type (ptr, limit)
 	break;
       }
     default:
-      fatal ("unrecognized signature string");
+      abort ();
     }
   return promote_type (type);
 }
@@ -498,7 +545,6 @@ parse_signature_string (sig_string, sig_length)
   const unsigned char *str = sig_string;
   const unsigned char *limit = str + sig_length;
 
-  push_obstacks (&permanent_obstack, &permanent_obstack);
   if (str < limit && str[0] == '(')
     {
       tree argtype_list = NULL_TREE;
@@ -509,7 +555,7 @@ parse_signature_string (sig_string, sig_length)
 	  argtype_list = tree_cons (NULL_TREE, argtype, argtype_list);
 	}
       if (str++, str >= limit)
-	fatal ("bad signature string");
+	abort ();
       result_type = parse_signature_type (&str, limit);
       argtype_list = chainon (nreverse (argtype_list), end_params_node);
       result_type = build_function_type (result_type, argtype_list);
@@ -518,7 +564,6 @@ parse_signature_string (sig_string, sig_length)
     result_type = parse_signature_type (&str, limit);
   if (str != limit)
     error ("junk at end of signature string");
-  pop_obstacks ();
   return result_type;
 }
 
@@ -529,7 +574,7 @@ parse_signature_string (sig_string, sig_length)
 tree
 get_type_from_signature (tree signature)
 {
-  unsigned char *sig = (unsigned char *) IDENTIFIER_POINTER (signature);
+  const unsigned char *sig = (const unsigned char *) IDENTIFIER_POINTER (signature);
   int len = IDENTIFIER_LENGTH (signature);
   tree type;
   /* Primitive types aren't cached. */
@@ -544,6 +589,15 @@ get_type_from_signature (tree signature)
   return type;
 }
 
+/* Ignore signature and always return null.  Used by has_method. */
+
+static tree
+build_null_signature (type)
+     tree type ATTRIBUTE_UNUSED;
+{
+  return NULL_TREE;
+}
+
 /* Return the signature string for the arguments of method type TYPE. */
 
 tree
@@ -579,15 +633,10 @@ build_java_signature (type)
      tree type;
 {
   tree sig, t;
-  push_obstacks (&permanent_obstack, &permanent_obstack);
   while (TREE_CODE (type) == POINTER_TYPE)
     type = TREE_TYPE (type);
-  if (TYPE_LANG_SPECIFIC (type) == NULL)
-    {
-      TYPE_LANG_SPECIFIC (type) = (struct lang_type *)
-	perm_calloc (1, sizeof (struct lang_type));
-    }
-  sig = TYPE_LANG_SPECIFIC (type)->signature;
+  MAYBE_CREATE_TYPE_TYPE_LANG_SPECIFIC (type);
+  sig = TYPE_SIGNATURE (type);
   if (sig == NULL_TREE)
     {
       char sg[2];
@@ -651,11 +700,10 @@ build_java_signature (type)
 	  break;
 	bad_type:
 	default:
-	  fatal ("internal error - build_java_signature passed invalid type");
+	  abort ();
 	}
-      TYPE_LANG_SPECIFIC (type)->signature = sig;
+      TYPE_SIGNATURE (type) = sig;
     }
-  pop_obstacks ();
   return sig;
 }
 
@@ -669,71 +717,144 @@ set_java_signature (type, sig)
   tree old_sig;
   while (TREE_CODE (type) == POINTER_TYPE)
     type = TREE_TYPE (type);
-  if (TYPE_LANG_SPECIFIC (type) == NULL)
-    {
-      TYPE_LANG_SPECIFIC (type) = (struct lang_type *)
-	perm_calloc (1, sizeof (struct lang_type));
-      
-    }
-  old_sig = TYPE_LANG_SPECIFIC (type)->signature;
+  MAYBE_CREATE_TYPE_TYPE_LANG_SPECIFIC (type);
+  old_sig = TYPE_SIGNATURE (type);
   if (old_sig != NULL_TREE && old_sig != sig)
-    fatal ("internal error - set_java_signature");
-  TYPE_LANG_SPECIFIC (type)->signature = sig;
+    abort ();
+  TYPE_SIGNATURE (type) = sig;
 #if 0 /* careful about METHOD_TYPE */
   if (IDENTIFIER_SIGNATURE_TYPE (sig) == NULL_TREE && TREE_PERMANENT (type))
     IDENTIFIER_SIGNATURE_TYPE (sig) = type;
 #endif
 }
 
-/* Search in class CLAS (and its superclasses) for a method
-   matching METHOD_NAME and argument signature METHOD_SIGNATURE.
-   Return a FUNCTION_DECL on success, or NULL_TREE if none found.
-   (Contrast lookup_java_method, which takes into account return type.) */
+/* Search in class SEARCHED_CLASS (and its superclasses) for a method
+   matching METHOD_NAME and signature SIGNATURE.  If SEARCHED_INTERFACE is
+   not NULL_TREE then first search its superinterfaces for a similar match.
+   Return the matched method DECL or NULL_TREE.  SIGNATURE_BUILDER is
+   used on method candidates to build their (sometimes partial)
+   signature.  */
 
 tree
-lookup_argument_method (clas, method_name, method_signature)
-     tree clas, method_name, method_signature;
+lookup_argument_method (searched_class, method_name, method_signature)
+     tree searched_class, method_name, method_signature;
 {
-  tree method;
-  while (clas != NULL_TREE)
-    {
-      for (method = TYPE_METHODS (clas);
-	   method != NULL_TREE;  method = TREE_CHAIN (method))
-	{
-	  tree method_sig = build_java_argument_signature (TREE_TYPE (method));
-	  tree name = DECL_NAME (method);
-	  if ((TREE_CODE (name) == EXPR_WITH_FILE_LOCATION ?
-	       EXPR_WFL_NODE (name) : name) == method_name 
-	      && method_sig == method_signature)
-	    return method;
-	}
-      clas = CLASSTYPE_SUPER (clas);
-    }
-  return NULL_TREE;
+  return lookup_do (searched_class, NULL_TREE, method_name, method_signature, 
+		    build_java_argument_signature);
 }
 
-/* Search in class CLAS (and its superclasses) for a method
-   matching METHOD_NAME and signature METHOD_SIGNATURE.
-   Return a FUNCTION_DECL on success, or NULL_TREE if none found.
-   (Contrast lookup_argument_method, which ignores return type.) */
+/* Search in class SEARCHED_CLASS (and its superclasses and
+   implemented interfaces) for a method matching METHOD_NAME and
+   argument signature METHOD_SIGNATURE.  Return a FUNCTION_DECL on
+   success, or NULL_TREE if none found.  (Contrast lookup_java_method,
+   which takes into account return type.) */
 
 tree
-lookup_java_method (clas, method_name, method_signature)
-     tree clas, method_name, method_signature;
+lookup_argument_method2 (searched_class, method_name, method_signature)
+     tree searched_class, method_name, method_signature;
+{
+  return lookup_do (CLASSTYPE_SUPER (searched_class), searched_class,
+		    method_name, method_signature, 
+		    build_java_argument_signature);
+}
+
+/* Search in class SEARCHED_CLASS (and its superclasses) for a method
+   matching METHOD_NAME and signature METHOD_SIGNATURE.  Return a
+   FUNCTION_DECL on success, or NULL_TREE if none found.  (Contrast
+   lookup_argument_method, which ignores return type.) If
+   SEARCHED_CLASS is an interface, search it too. */
+
+tree
+lookup_java_method (searched_class, method_name, method_signature)
+     tree searched_class, method_name, method_signature;
+{
+  tree searched_interface;
+  
+  /* If this class is an interface class, search its superinterfaces
+   * first. A superinterface is not an interface's superclass: a super
+   * interface is implemented by the interface.  */
+
+  searched_interface = (CLASS_INTERFACE (TYPE_NAME (searched_class)) ?
+			searched_class : NULL_TREE);
+  return lookup_do (searched_class, searched_interface, method_name, 
+		    method_signature, build_java_signature);
+}
+
+/* Return true iff CLASS (or its ancestors) has a method METHOD_NAME. */
+
+int
+has_method (class, method_name)
+     tree class;
+     tree method_name;
+{
+  return lookup_do (class, class,  method_name,
+		    NULL_TREE, build_null_signature) != NULL_TREE;
+}
+
+/* Search in class SEARCHED_CLASS (and its superclasses) for a method
+   matching METHOD_NAME and signature SIGNATURE.  Also search in
+   SEARCHED_INTERFACE (and its superinterfaces) for a similar match.
+   Return the matched method DECL or NULL_TREE.  SIGNATURE_BUILDER is
+   used on method candidates to build their (sometimes partial)
+   signature.  */
+
+static tree
+lookup_do (searched_class, searched_interface, method_name, signature, signature_builder)
+     tree searched_class, searched_interface, method_name, signature;
+     tree (*signature_builder) PARAMS ((tree));
 {
   tree method;
-  while (clas != NULL_TREE)
+  
+  if (searched_interface)
     {
-      for (method = TYPE_METHODS (clas);
+      int i;
+      int interface_len = 
+	TREE_VEC_LENGTH (TYPE_BINFO_BASETYPES (searched_interface)) - 1;
+
+      for (i = interface_len; i > 0; i--)
+       {
+         tree child = 
+	   TREE_VEC_ELT (TYPE_BINFO_BASETYPES (searched_interface), i);
+         tree iclass = BINFO_TYPE (child);
+
+         /* If the superinterface hasn't been loaded yet, do so now.  */
+	 if (CLASS_FROM_SOURCE_P (iclass))
+	   safe_layout_class (iclass);
+	 else if (!CLASS_LOADED_P (iclass))
+	   load_class (iclass, 1);
+
+         for (method = TYPE_METHODS (iclass);
+              method != NULL_TREE;  method = TREE_CHAIN (method))
+           {
+             tree method_sig = (*signature_builder) (TREE_TYPE (method));
+
+	     if (DECL_NAME (method) == method_name && method_sig == signature)
+               return method;
+           }
+
+         /* it could be defined in a supersuperinterface */
+         if (CLASS_INTERFACE (TYPE_NAME (iclass)))
+           {
+             method = lookup_do (iclass, iclass, method_name, 
+				 signature, signature_builder);
+             if (method != NULL_TREE) 
+	       return method;
+           }
+       }
+    }
+
+  while (searched_class != NULL_TREE)
+    {
+      for (method = TYPE_METHODS (searched_class);
 	   method != NULL_TREE;  method = TREE_CHAIN (method))
 	{
-	  tree method_sig = build_java_signature (TREE_TYPE (method));
-	  if (DECL_NAME (method) == method_name 
-	      && method_sig == method_signature)
+	  tree method_sig = (*signature_builder) (TREE_TYPE (method));
+	  if (DECL_NAME (method) == method_name && method_sig == signature)
 	    return method;
 	}
-      clas = CLASSTYPE_SUPER (clas);
+      searched_class = CLASSTYPE_SUPER (searched_class);
     }
+
   return NULL_TREE;
 }
 
@@ -756,7 +877,7 @@ lookup_java_constructor (clas, method_signature)
 
 /* Return a type which is the Binary Numeric Promotion of the pair T1,
    T2 and convert EXP1 and/or EXP2. See 5.6.2 Binary Numeric
-   Promotion. It assumes that both T1 and T2 are elligible to BNP. */
+   Promotion. It assumes that both T1 and T2 are eligible to BNP. */
 
 tree
 binary_numeric_promotion (t1, t2, exp1, exp2)
